var keychain;
var baseUrl;
var keychainFolder, dataFolder, styleFolder;
var current_profile = null;
var devmode;

function mainPageDidFinishLoading() {
	try {
		setup();
		keychain = new Keychain();
		showLockedScreen();
	} catch(e) {
		showFatalException(e);
	}
}

function showLockedScreen() {
	try {
		var locked = styleFolder + "/html/login.html";
		loadFile(locked, loginHtmlDidFinishLoading);
	} catch(e) {
		showFatalException(e);
	}
}

function loginHtmlDidFinishLoading(txt) {
	if (!txt || txt == "") {
		var file = styleFolder + "/html/login.html";
		showFatalMessage("Problem loading 1Password data file", "<p>A <a href='" + file + "'>key data file</a> could not be loaded and 1PasswordAnywhere cannot continue without it.</p><p>Please see <a href='http://help.agile.ws/1Password3/1passwordanywhere_troubleshooting.html'>this help guide</a> for troubleshooting tips.</p>");
		return;
	}
	
	try {
		$('mainBody').innerHTML = txt;
	
		// Todo: allProfiles should be dynamically loaded
		var allProfiles = ["default"]; 
		selectProfile(allProfiles[0]); 
		$("masterPassword").focus();
	} catch(e) {
		showFatalException(e);
	}
}

function verifyPassword(password) {
	try {
		if (!keychain.encryptionKeysLoaded) {
			encryptionKeysCouldNotLoad();
			return;
		}
		
		if (keychain.verifyPassword(password)) {
			var opened = styleFolder + "/html/main.html";
			loadFile(opened, unlockedHtmlDidFinishLoading);
		}
		else {
			login_failed();
		}
	} catch(e) {
		showFatalException(e);
	}
}

function unlockedHtmlDidFinishLoading(txt) {
	if (!txt || txt == "") {
		var file = styleFolder + "/html/main.html";
		alert("Unable to load the '" + file + "' file. This is a critical file and 1PasswordAnywhere cannot proceed without it.");
		return;
	}
	
	try {
		$('mainBody').innerHTML = txt;
		loadFile(fullKeychainFilePath('contents.js'), profileContentsDidFinishLoading);
	} catch(e) {
		showFatalException(e);
	}
}

function encryptionKeysCouldNotLoad() {
	var keysFilePath = fullKeychainFilePath('encryptionKeys.js');
	showFatalMessage("Problem loading 1Password data file", "<p>A <a target='_blank' href='" + keysFilePath + "'>key data file</a> could not be loaded and 1PasswordAnywhere cannot continue without it.</p><p>Please see <a target='_blank'  href='http://help.agile.ws/1Password3/1passwordanywhere_troubleshooting.html'>this help guide</a> for troubleshooting tips.</p>");
}

function encryptionKeysDidFinishLoading(json) {
	if (!json || json == "") {
		encryptionKeysCouldNotLoad();
		return;
	}
	
	try {
		var keys;
		try {
			json = "(" + json + ")";
			keys = eval(json);
		}
		catch(e) {
			showFatalMessage("Problem parsing 1Password data file", "<p>There was a problem parsing the data contained in encryptionKeys.js.</p><p>Please <a href='mailto:1PAnywhere@agile.ws?subject=1PasswordAnywhere%20problem%20parsing%20encryptionKeys.js&body=Please let us know what you were doing when this problem occurred.'>report this problem</a> to the Agile team.</p>")
		}
		
		keychain.setEncryptionKeys(keys);
	} catch(e) {
		showFatalException(e);
	}
}

function selectProfile(profile_name) {
	try {
		top.current_profile = profile_name;
		loadFile(fullKeychainFilePath('encryptionKeys.js'), encryptionKeysDidFinishLoading);
	} catch(e) {
		showFatalException(e);
	}
}

function setup() {
	try {
		_setup();
	} catch(e) {
		showFatalException(e);
	}
}

function _setup() {
	baseUrl = window.location.href.substring(0, window.location.href.indexOf("1Password.html"));
	keychainFolder = baseUrl;

	var parameters = window.location.search.substring(1).split("&");

	for (var i=0; i < parameters.length; i++) {
		var nvp = parameters[i].split("=");
		var name = nvp[0], value = nvp[1];

		if (name == "keychainFolder") {
			keychainFolder = value;
			if (keychainFolder[keychainFolder.length-1] != "/") keychainFolder += "/";
		}
		if (name == "keychainFolder" && navigator && navigator.userAgent.match("Firefox") ) {
			alert("FATAL: The keychainFolder parameter is NOT supported in Firefox. Accessing files from outside the folder that opens you is against the mozilla security model unless you set security.fileuri.strict_origin_policy to false.");
		}
		if (name == "devmode") {
			devmode = true;
		}
	}
	
	if (devmode) {
		styleFolder = "./style";
	}
	else {
		styleFolder = keychainFolder + "/style";
	}
	dataFolder = keychainFolder + "/data";
}

function logout(autologout) {
	try {
		// Not really needed but want to be 100% sure decrypted data is cleared
		clearEntryDetails();
		keychain.lock();
		
		// TODO -- would be nice to show autologout message to user
		showLockedScreen();
    } catch(e) {
        showFatalException(e);
    }
}

function setFocus()
{
	$('password').focus();
}

function fullKeychainFilePath(filename) {
	return top.keychainFolder + "data/" + top.current_profile + "/" + filename;
}

function loadFile(file, onSuccess) {
	enableFirefoxPrivileges();
	new Ajax.Request(file,
	  {
		method: 'get',
	    onSuccess: function(transport){ onSuccess(transport.responseText);},
		onFailure: function() { alert('A problem occurred when loading the "' + file + '" file.'); }, 
	  });
}

function enableFirefoxPrivileges() {
	// request local file read permission in Firefox
	try {
			try {
				// Make Safari bomb out
				netscape;
				netscape.security.PrivilegeManager.enablePrivilege("UniversalFileRead");
			}
			catch(e) {}
	} catch (e) {
			alert("Permission to read file was denied. If in Firefox you must allow this script to load local files.");
	}
}

function showFatalException(exc) {
	try {
		var error_desc = "Exception was nil."
		if (exc) {
			var file = exc.sourceURL;
			if (!file) file = exc.fileName;
			if (file && file.match(/^.*\/(.*)\?ts=.*$/)) {
				file = RegExp.$1;
			}
			error_desc = exc.name + " occurred in " + file + " on line #" + exc.line + ":\n" + exc.message;
		}

		if (OPANYWHERE_VERSION) error_desc += "\nYou are running 1PasswordAnywhere version #" + OPANYWHERE_VERSION;
		
		var html_msg = "<p>" + error_desc.gsub(/\n/, "<br/>") + "</p><p>Please <a href='mailto:1PAnywhere@agile.ws?subject=1PasswordAnywhere%20Problem&body=Please let us know what you were doing when this problem occurred.%0A%0AError details:%0A" + escape(error_desc) + "'>report this error</a> so Agile can investigate.</p>";
		_showMessageBox("1PasswordAnywhere experienced an error", html_msg, 'messageBox_error', {'fatal':true});
	}
	catch(e) {
		var msg = 'An exception occurred during error processing! Please report this issue to Agile at 1PAnywhere@agile.ws along with a description of what you were doing when this error happened.';
		if (OPANYWHERE_VERSION) msg += ' You are running version #' + OPANYWHERE_VERSION;
		alert(msg);
	}
}

function showMessage(title, message) {
	_showMessageBox(title, message, 'messageBox_error', {});
}

function showFatalMessage(title, message) {
	_showMessageBox(title, message, 'messageBox_error', { 'fatal' : true });
}

function _showMessageBox (title, message, cssClass, options) {
	var help_url = options['help_url'];
	if (help_url) {
		$("messageBox_help").href = help_url;
	}
	else {
		$("messageBox_help").hide();
	}

	$('messageBox').writeAttribute('class', cssClass);
	$('messageBox_title').innerHTML = title;
	$('messageBox_message').innerHTML = message;
		
	// Effect.Appear
	$('messageBox_wrapper').show();
	Effect.Appear('messageBox');
	
	var fatal = options['fatal'];
	if (fatal) {
		$("messageBox_dismiss").hide();
	}
	else {
		Event.observe('messageBox_dismiss', 'click', function(){
			//Dismiss message box.
			Effect.Fade('messageBox_wrapper', {queue:"end"});
			Effect.Fade('messageBox', {queue:"end"});

			$('messageBox_dismiss').stopObserving('click');
		});
		Event.observe('messageBox_wrapper', 'click', function(){
			//Dismiss message box.
			Effect.Fade('messageBox_wrapper', {queue:"end"});
			Effect.Fade('messageBox', {queue:"end"});

			$('messageBox_dismiss').stopObserving('click');
		});		
	}
}

function login_failed()
{
	try {
		$('passwordLabel').setAttribute('class', 'error');
		$('masterPassword').setAttribute('class', 'error');
		$('masterPassword').value = "";
	
		var hint = fullKeychainFilePath(".password.hint");
		new Ajax.Request(hint, {
			onComplete: function(transport){showPasswordHintIfAvailable(transport)},
		  });
		
		Effect.Shake('passwordLabel');
	}
	catch(e) {
		showFatalException(e);
	}
}

function showPasswordHintIfAvailable(transport) {
	if (transport.responseText.length != 0) {
		$('passwordError').style.display = '';
		$('masterPasswordHint').innerHTML = transport.responseText.escapeHTML();
	}
}